@@ -9,9 +9,11 @@ import pytest
 
 import tomlkit
 
+from tests.util import assert_is_ppo
 from tomlkit import parse
 from tomlkit import ws
 from tomlkit._utils import _utc
+from tomlkit.api import document
 from tomlkit.exceptions import NonExistentKey
 
 
@@ -153,6 +155,20 @@ name = "bar"
     assert "tool" in d
 
 
+def test_toml_document_unwrap():
+    content = """[tool.poetry]
+name = "foo"
+"""
+
+    doc = parse(content)
+    unwrapped = doc.unwrap()
+    assert_is_ppo(unwrapped, dict)
+    assert_is_ppo(list(unwrapped.keys())[0], str)
+    assert_is_ppo(unwrapped["tool"], dict)
+    assert_is_ppo(list(unwrapped["tool"].keys())[0], str)
+    assert_is_ppo(unwrapped["tool"]["poetry"]["name"], str)
+
+
 def test_toml_document_with_dotted_keys(example):
     content = example("0.5.0")
 
@@ -444,7 +460,8 @@ name = "foo"
     dev_dependencies["baz"]["source"] = "other"
 
     assert (
-        """\
+        doc.as_string()
+        == """\
 [tool.poetry]
 name = "foo"
 
@@ -455,7 +472,6 @@ bar = {version = "^3.0", source = "remote"}
 [tool.poetry.dev-dependencies]
 baz = {version = "^4.0", source = "other"}
 """
-        == doc.as_string()
     )
 
 
@@ -491,7 +507,7 @@ key = "value"
     doc = parse(content)
     doc["a"]["a"]["key"] = "new_value"
 
-    assert "new_value" == doc["a"]["a"]["key"]
+    assert doc["a"]["a"]["key"] == "new_value"
 
     expected = """
 [a.a]
@@ -571,18 +587,18 @@ key = "value"
     table = doc["a"]["a"]
     assert "key" in table
     assert "c" in table
-    assert "value" == table.get("key")
+    assert table.get("key") == "value"
     assert {} == table.get("c")
     assert table.get("d") is None
-    assert "foo" == table.get("d", "foo")
+    assert table.get("d", "foo") == "foo"
 
-    assert "bar" == table.setdefault("d", "bar")
-    assert "bar" == table["d"]
+    assert table.setdefault("d", "bar") == "bar"
+    assert table["d"] == "bar"
 
-    assert "value" == table.pop("key")
+    assert table.pop("key") == "value"
     assert "key" not in table
 
-    assert "baz" == table.pop("missing", default="baz")
+    assert table.pop("missing", default="baz") == "baz"
 
     with pytest.raises(KeyError):
         table.pop("missing")
@@ -612,7 +628,7 @@ a = "b"
     constraint["version"] = "^1.0"
     doc["tool"]["poetry"]["dependencies"]["bar"] = constraint
 
-    assert "^1.0" == doc["tool"]["poetry"]["dependencies"]["bar"]["version"]
+    assert doc["tool"]["poetry"]["dependencies"]["bar"]["version"] == "^1.0"
 
     expected = """
 [tool.poetry]
@@ -635,6 +651,31 @@ a = "b"
     assert expected == doc.as_string()
 
 
+def test_remove_from_out_of_order_table():
+    content = """[a]
+x = 1
+
+[c]
+z = 3
+
+[a.b]
+y = 2
+"""
+    document = parse(content)
+    del document["a"]["b"]
+    assert (
+        document.as_string()
+        == """[a]
+x = 1
+
+[c]
+z = 3
+
+"""
+    )
+    assert json.dumps(document) == '{"a": {"x": 1}, "c": {"z": 3}}'
+
+
 def test_updating_nested_value_keeps_correct_indent():
     content = """
 [Key1]
@@ -927,3 +968,119 @@ def test_pop_add_whitespace_and_insert_table_work_togheter():
     assert out["d"] == 4
     assert "d" not in out["e"]
     assert text == dedent(expected)
+
+
+def test_add_newline_before_super_table():
+    doc = document()
+    doc["a"] = 1
+    doc["b"] = {"c": {}}
+    doc["d"] = {"e": {}}
+    expected = """\
+    a = 1
+
+    [b.c]
+
+    [d.e]
+    """
+    assert doc.as_string() == dedent(expected)
+
+
+def test_remove_item_from_super_table():
+    content = """\
+    [hello.one]
+    a = 1
+
+    [hello.two]
+    b = 1
+    """
+    doc = parse(dedent(content))
+    del doc["hello"]["two"]
+    expected = """\
+    [hello.one]
+    a = 1
+
+    """
+    assert doc.as_string() == dedent(expected)
+
+
+def test_nested_table_update_display_name():
+    content = """\
+    [parent]
+
+    [parent.foo]
+    x = 1
+    """
+
+    doc = parse(dedent(content))
+    sub = """\
+    [foo]
+    y = 2
+
+    [bar]
+    z = 3
+    """
+    doc["parent"].update(parse(dedent(sub)))
+    expected = """\
+    [parent]
+
+    [parent.foo]
+    y = 2
+
+    [parent.bar]
+    z = 3
+    """
+    assert doc.as_string() == dedent(expected)
+
+
+def test_build_table_with_dotted_key():
+    doc = tomlkit.document()
+    data = {
+        "a.b.c": 1,
+        "a.b.d": 2,
+        "a": {"c": {"foo": "bar"}},
+        "a.d.e": 3,
+    }
+
+    for key, value in data.items():
+        if "." not in key:
+            doc.append(key, value)
+        else:
+            doc.append(tomlkit.key(key.split(".")), value)
+
+    expected = """\
+a.b.c = 1
+a.b.d = 2
+a.d.e = 3
+
+[a.c]
+foo = "bar"
+"""
+    assert doc.as_string() == expected
+    assert json.loads(json.dumps(doc)) == {
+        "a": {"b": {"c": 1, "d": 2}, "d": {"e": 3}, "c": {"foo": "bar"}}
+    }
+
+
+def test_parse_subtables_no_extra_indent():
+    expected = """\
+[a]
+    [a.b.c]
+        foo = 1
+
+    [a.b.d]
+        bar = 2
+"""
+    doc = parse(expected)
+    assert doc.as_string() == expected
+
+
+def test_item_preserves_the_order():
+    t = tomlkit.inline_table()
+    t.update({"a": 1, "b": 2})
+    doc = {"name": "foo", "table": t, "age": 42}
+    expected = """\
+name = "foo"
+table = {a = 1, b = 2}
+age = 42
+"""
+    assert tomlkit.dumps(doc) == expected
